@@ -55,6 +55,8 @@
 #include <pty.h>
 #endif
 
+#define A_DECL(type, var) type var##1, var##2, *var = _THUNK_PTR_STRUCT_OK(&var##1) ? &var##1 : &var##2
+
 #endif
 
 #ifdef OS2
@@ -64,41 +66,64 @@
  * TCP/IP 4.0 returns EFAULT if we do I/O to/from high address - we test for
  * EFAULT and retry with a bounce buffer.  */
 
-#define BOUNCE_BUFFER_SIZE	256
+#define BOUNCE_BUFFER_SIZE	3584
 
 int bounced_read(int fd, void *buf, size_t size)
 {
 	unsigned char *bounce_buffer;
-	size_t xsiz;
 	int r;
-	if (fd < 3 && (unsigned long)buf + size > 0x20000000) goto bounce;
+
+	if (fd < 3 && (my_uintptr_t)buf + size >= 0x20000000UL) goto bounce;
 	r = _read(fd, buf, size);
 	if (r == -1 && errno == EFAULT) goto bounce;
 	return r;
-	bounce:
-	xsiz = size > BOUNCE_BUFFER_SIZE ? BOUNCE_BUFFER_SIZE : size;
-	bounce_buffer = alloca(xsiz);
-	r = _read(fd, bounce_buffer, xsiz);
+
+bounce:
+	bounce_buffer = NULL;
+	if (size > BOUNCE_BUFFER_SIZE) {
+		bounce_buffer = os2_orig_malloc(size);
+		if (!bounce_buffer)
+			size = BOUNCE_BUFFER_SIZE;
+	}
+	if (!bounce_buffer)
+		bounce_buffer = alloca(size);
+	r = _read(fd, bounce_buffer, size);
 	if (r > 0) memcpy(buf, bounce_buffer, r);
+	if (size > BOUNCE_BUFFER_SIZE)
+		free(bounce_buffer);
 	return r;
 }
 
 int bounced_write(int fd, const void *buf, size_t size)
 {
 	unsigned char *bounce_buffer;
-	size_t xsiz;
 	int r;
-	if (fd < 3 && (unsigned long)buf + size > 0x20000000) goto bounce;
+
+	if (fd < 3 && (my_uintptr_t)buf + size >= 0x20000000UL) goto bounce;
 	r = _write(fd, buf, size);
 	if (r == -1 && errno == EFAULT) goto bounce;
 	return r;
-	bounce:
-	xsiz = size > BOUNCE_BUFFER_SIZE ? BOUNCE_BUFFER_SIZE : size;
-	bounce_buffer = alloca(xsiz);
-	memcpy(bounce_buffer, buf, xsiz);
-	return _write(fd, bounce_buffer, xsiz);
+
+bounce:
+	bounce_buffer = NULL;
+	if (size > BOUNCE_BUFFER_SIZE) {
+		bounce_buffer = os2_orig_malloc(size);
+		if (!bounce_buffer)
+			size = BOUNCE_BUFFER_SIZE;
+	}
+	if (!bounce_buffer)
+		bounce_buffer = alloca(size);
+	memcpy(bounce_buffer, buf, size);
+	r = _write(fd, bounce_buffer, size);
+	if (size > BOUNCE_BUFFER_SIZE)
+		free(bounce_buffer);
+	return r;
 }
 
+void portable_sleep(unsigned msec)
+{
+	DosSleep(msec);
+}
 #endif
 
 #ifdef OS2_ADVANCED_HEAP
@@ -115,68 +140,83 @@ unsigned long blocks_requested = 0;
 static int dosallocmem_attrib = PAG_READ | PAG_WRITE | PAG_COMMIT;
 
 #define HEAP_ALIGN		0x10000
-#define HEAP_PAD		2
-#define HEAP_MAXPAD		0x1000000
-#define HEAP_IMMEDIATE_RELEASE	0x100000
+#define HEAP_MAX_ALIGN		0x20000
+
+void virtual_free(void *ptr, size_t len)
+{
+	int rc;
+	rc = DosFreeMem(ptr);
+	/*fprintf(stderr, "heap free %p -> %d\n", ptr, rc);*/
+	if (rc)
+		fatal_exit("DosFreeMem failed: %d", rc);
+	if (len & 4095)
+		len = (len | 4095) + 1;
+	mem_requested -= len;
+	blocks_requested--;
+}
 
 static void heap_release(Heap_t h, void *ptr, size_t len)
 {
+	virtual_free(ptr, len);
+}
+
+void *virtual_alloc(size_t len)
+{
+	void *result;
 	int rc;
-	mem_requested -= ((len | 4095) + 1);
-	blocks_requested--;
-	rc = DosFreeMem(ptr);
-	/*fprintf(stderr, "heap free %p -> %d\n", ptr, rc);*/
-	if (rc) {
-		fatal_exit("DosFreeMem failed: %d", rc);
+	rc = DosAllocMem(&result, len, dosallocmem_attrib);
+	if (rc)
+		return NULL;
+	/*
+	 * Hitting the shared arena has a negative impact on the whole
+	 * system. Therefore, we fake failure (so that Links frees
+	 * some caches) and try allocating near the shared arena only
+	 * as a last resort.
+	 */
+	if ((unsigned long)result >= 0x12000000 &&
+	    (unsigned long)result < 0x20000000) {
+		if (!malloc_try_hard) {
+			heap_release(NULL, result, len);
+			return NULL;
+		}
 	}
+	if (len & 4095)
+		len = (len | 4095) + 1;
+	mem_requested += len;
+	blocks_requested++;
+	return result;
 }
 
 static void *heap_alloc(Heap_t h, size_t *size, int *pclean)
 {
 	void *result;
-	int rc;
 	/* If we rounded up to page size, EMX would join all allocations
 	 * to one segment and refuse to free memory. So round up to
 	 * page size - 1 */
 	size_t real_size = *size;
-	if (real_size < HEAP_MAXPAD / HEAP_PAD) {
-		real_size *= HEAP_PAD;
+	if (real_size < HEAP_MAX_ALIGN) {
 		real_size = real_size | (HEAP_ALIGN - 1);
 	} else {
 		real_size |= 1;
 	}
-	rc = DosAllocMem(&result, real_size, dosallocmem_attrib);
-	/*fprintf(stderr, "heap alloc %d -> %p, %d\n", *size, result, rc);*/
-	if (!rc) {
-		/*
-		 * Hitting the shared arena has a negative impact on the whole
-		 * system. Therefore, we fake failure (so that Links frees
-		 * some caches) and try allocating near the shared arena only
-		 * as a last resort.
-		 */
-		if ((unsigned long)result >= 0x12000000 &&
-		    (unsigned long)result < 0x20000000) {
-			if (!malloc_try_hard) {
-				heap_release(NULL, result, real_size);
-				return NULL;
-			}
-		}
+	result = virtual_alloc(real_size);
+	/*fprintf(stderr, "heap alloc %d,%d -> %p\n", *size, real_size, result);*/
+	if (result) {
 		*size = real_size;
 		*pclean = _BLOCK_CLEAN;
-		mem_requested += ((real_size | 4095) + 1);
-		blocks_requested++;
-		return result;
-	} else {
-		return NULL;
 	}
+	return result;
 }
 
+static Heap_t original_heap = NULL;
+
 static void init_os2_heap(void)
 {
 	Heap_t new_heap;
 	size_t init_size = _HEAP_MIN_SIZE;
 	void *init_mem;
 	int init_clean;
+
 	dosallocmem_attrib |= OBJ_ANY;
 	init_mem = heap_alloc(NULL, &init_size, &init_clean);
 	if (!init_mem) {
@@ -199,15 +239,21 @@ static void init_os2_heap(void)
 #endif
 		return;
 	}
-	_udefault(new_heap);
+	if (dosallocmem_attrib & OBJ_ANY) {
+		/* call malloc to initialize the default heap */
+		void *p = malloc(1);
+		if (p) free(p);
+		original_heap = _udefault(new_heap);
+	} else {
+		_udefault(new_heap);
+	}
 }
 
-void mem_freed_large(size_t size)
+void *os2_orig_malloc(size_t len)
 {
-#if defined(HAVE__HEAPMIN)
-	if (size >= HEAP_IMMEDIATE_RELEASE)
-		_heapmin();
-#endif
+	if (original_heap)
+		return _umalloc(original_heap, len);
+	return malloc(len);
 }
 
 #endif
@@ -222,7 +268,7 @@ static HSWITCH os2_switchhandle = NULLHANDLE;
 #ifdef HAVE_SYS_FMUTEX_H
 static _fmutex fd_mutex;
 #endif
-#elif defined(HAVE_PTHREADS)
+#elif defined(HAVE_PTHREADS) && !defined(OPENVMS)
 static pthread_mutex_t pth_mutex;
 static void fd_lock(void);
 static void fd_unlock(void);
@@ -235,6 +281,22 @@ static void fd_init(void)
 }
 #endif
 
+int page_size = 4096;
+
+void init_page_size(void)
+{
+	long getpg = -1;
+#if defined(HAVE_SYSCONF) && defined(_SC_PAGESIZE)
+	if (getpg < 0)
+		EINTRLOOP(getpg, sysconf(_SC_PAGESIZE));
+#endif
+#ifdef HAVE_GETPAGESIZE
+	if (getpg < 0)
+		getpg = getpagesize();
+#endif
+	if (getpg > 0 && !(getpg & (getpg - 1))) page_size = (int)getpg;
+}
+
 #if !defined(OPENVMS) && !defined(DOS)
 void init_os(void)
 {
@@ -258,7 +320,7 @@ void init_os(void)
 	if (_fmutex_create(&fd_mutex, 0))
 		fatal_exit("failed to create fd mutex");
 #endif
-#elif defined(HAVE_PTHREADS)
+#elif defined(HAVE_PTHREADS) && !defined(OPENVMS)
 	{
 		int r;
 		fd_init();
@@ -293,7 +355,7 @@ void init_os(void)
 	if (getppid() == 1) {
 		int i;
 		for (i = 1; i < g_argc; i++) {
-			unsigned char *a = g_argv[i];
+			unsigned char *a = cast_uchar g_argv[i];
 			int l = (int)strlen(cast_const_char a);
 			if (l >= 3 && a[0] == '"' && a[l - 1] == '"') {
 				unsigned char *a2 = cast_uchar strdup(cast_const_char (a + 1));
@@ -306,13 +368,19 @@ void init_os(void)
 						*q++ = *p;
 					}
 					*q = 0;
-					g_argv[i] = a2;
+					g_argv[i] = cast_char a2;
 				}
 			}
 		}
 	}
 #endif
 #endif
+#if defined(BEOS) || defined(HAIKU)
+	if (getenv("TERM") == NULL) {
+		/* probably launched from Tracker or Deskbar, force graphics mode */
+		force_g = 1;
+	}
+#endif
 }
 #endif
 
@@ -396,7 +464,9 @@ void ignore_signals(void)
 #ifdef SIGCHLD
 	do_signal(SIGCHLD, SIG_IGN);
 #endif
-	do_signal(SIGINT, SIG_IGN);
+#ifdef SIGWINCH
+	do_signal(SIGWINCH, SIG_IGN);
+#endif
 #endif
 }
 
@@ -454,71 +524,137 @@ void os_free_clipboard(void)
 
 /* Terminal size */
 
-#if defined(WIN) || defined(OPENVMS)
+#if defined(UNIX) || defined(OS2) || defined(WIN) || defined(INTERIX) || defined(BEOS) || defined(RISCOS) || defined(ATHEOS) || defined(SPAD) || defined(OPENVMS) || defined(HAIKU)
+static void (*terminal_resize_callback)(int, int);
+#endif
+
+#if (defined(OS2) && defined(X2)) || defined(WIN) || defined(OPENVMS)
+
+#define TERMINAL_SIZE_POLLING
 
 /* Cygwin has a bug and loses SIGWINCH sometimes, so poll it */
 
-static void winch_thread(void *p, int l)
+static struct timer *terminal_resize_timer = NULL;
+static int old_xsize, old_ysize;
+
+static void terminal_resize_fn(void *p)
 {
-	static int old_xsize, old_ysize;
-	static int cur_xsize, cur_ysize;
-	if (get_terminal_size(1, &old_xsize, &old_ysize)) return;
-	while (1) {
-		if (get_terminal_size(1, &cur_xsize, &cur_ysize)) return;
-		if ((old_xsize != cur_xsize) || (old_ysize != cur_ysize)) {
-			int rr;
-			old_xsize = cur_xsize;
-			old_ysize = cur_ysize;
-			EINTRLOOP(rr, raise(SIGWINCH));
-		}
-		sleep(1);
+	int cur_xsize, cur_ysize;
+	terminal_resize_timer = install_timer(TERMINAL_POLL_TIMEOUT, terminal_resize_fn, NULL);
+	get_terminal_size(&cur_xsize, &cur_ysize);
+	if ((old_xsize != cur_xsize) || (old_ysize != cur_ysize)) {
+		old_xsize = cur_xsize;
+		old_ysize = cur_ysize;
+		terminal_resize_callback(cur_xsize, cur_ysize);
 	}
 }
 
-static void terminal_resize_poll(void)
+static void terminal_resize_poll(int x, int y)
 {
-	static int winch_thread_running = 0;
-	if (!winch_thread_running) {
-		if (start_thread(winch_thread, NULL, 0, 0) >= 0)
-			winch_thread_running = 1;
-	}
+	if (terminal_resize_timer)
+		internal_error("terminal_resize_poll: timer already active");
+	old_xsize = x;
+	old_ysize = y;
+#ifdef OS2
+	if (!is_xterm())
+		return;
+#endif
+	terminal_resize_timer = install_timer(TERMINAL_POLL_TIMEOUT, terminal_resize_fn, NULL);
 }
 
 #endif
 
-#if defined(UNIX) || defined(WIN) || defined(INTERIX) || defined(BEOS) || defined(RISCOS) || defined(ATHEOS) || defined(SPAD) || defined(OPENVMS)
+#if defined(UNIX) || defined(OS2) || defined(WIN) || defined(INTERIX) || defined(BEOS) || defined(RISCOS) || defined(ATHEOS) || defined(SPAD) || defined(OPENVMS) || defined(HAIKU)
 
-#ifdef SIGWINCH
+#if defined(SIGWINCH) && !defined(OS2) && !defined(OPENVMS)
 static void sigwinch(void *s)
 {
-	((void (*)(void))s)();
+	int cur_xsize, cur_ysize;
+	get_terminal_size(&cur_xsize, &cur_ysize);
+#ifdef TERMINAL_SIZE_POLLING
+	old_xsize = cur_xsize;
+	old_ysize = cur_ysize;
+#endif
+	terminal_resize_callback(cur_xsize, cur_ysize);
 }
 #endif
 
-void handle_terminal_resize(int fd, void (*fn)(void))
+void handle_terminal_resize(void (*fn)(int, int), int *x, int *y)
 {
-#ifdef SIGWINCH
-	install_signal_handler(SIGWINCH, sigwinch, (void *)fn, 0);
+	terminal_resize_callback = fn;
+	get_terminal_size(x, y);
+#if defined(SIGWINCH) && !defined(OS2) && !defined(OPENVMS)
+	install_signal_handler(SIGWINCH, sigwinch, NULL, 0);
 #endif
-#ifdef WIN
-	terminal_resize_poll();
-#endif
-#ifdef OPENVMS
-	if (is_xterm())
-		terminal_resize_poll();
+#ifdef TERMINAL_SIZE_POLLING
+	terminal_resize_poll(*x, *y);
 #endif
 }
 
-void unhandle_terminal_resize(int fd)
+void unhandle_terminal_resize(void)
 {
-#ifdef SIGWINCH
+#if defined(SIGWINCH) && !defined(OS2) && !defined(OPENVMS)
 	install_signal_handler(SIGWINCH, NULL, NULL, 0);
 #endif
+#ifdef TERMINAL_SIZE_POLLING
+	if (terminal_resize_timer) {
+		kill_timer(terminal_resize_timer);
+		terminal_resize_timer = NULL;
+	}
+#endif
 }
 
-#if !defined(OPENVMS)
+#if defined(OS2)
 
-int get_terminal_size(int fd, int *x, int *y)
+int is_xterm(void)
+{
+	static int xt = -1;
+	if (xt == -1) xt = !!getenv("WINDOWID");
+	return xt;
+}
+
+void get_terminal_size(int *x, int *y)
+{
+	if (is_xterm()) {
+#ifdef X2
+		int arc;
+		struct winsize win;
+
+		/* fd = STDIN_FILENO; */
+		arc = ptioctl(1, TIOCGWINSZ, &win);
+		if (arc) {
+			*x = 80;
+			*y = 24;
+			return;
+		}
+		*y = win.ws_row;
+		*x = win.ws_col;
+		goto set_default;
+#else
+		*x = 80; *y = 24;
+#endif
+	} else {
+		int a[2] = { 0, 0 };
+		_scrsize(a);
+		*x = a[0];
+		*y = a[1];
+#ifdef X2
+		set_default:
+#endif
+		if (*x == 0) {
+			*x = get_e("COLUMNS");
+			if (*x == 0) *x = 80;
+		}
+		if (*y == 0) {
+			*y = get_e("LINES");
+			if (*y == 0) *y = 24;
+		}
+	}
+}
+
+#elif !defined(OPENVMS)
+
+void get_terminal_size(int *x, int *y)
 {
 	int rs = -1;
 #ifdef TIOCGWINSZ
@@ -543,123 +679,10 @@ int get_terminal_size(int fd, int *x, int *y)
 		) && !(*y = get_e(cast_uchar "LINES"))) {
 		*y = 24;
 	}
-	return 0;
 }
 
 #endif
 
-#elif defined(OS2)
-
-#define A_DECL(type, var) type var##1, var##2, *var = _THUNK_PTR_STRUCT_OK(&var##1) ? &var##1 : &var##2
-
-int is_xterm(void)
-{
-	static int xt = -1;
-	if (xt == -1) xt = !!getenv("WINDOWID");
-	return xt;
-}
-
-static int winch_pipe[2];
-static unsigned char winch_thread_running = 0;
-
-#define WINCH_SLEEPTIME 500 /* time in ms for winch thread to sleep */
-
-static void winch_thread(void *arg)
-{
-	/* A thread which regularly checks whether the size of
-	   window has changed. Then raise SIGWINCH or notifiy
-	   the thread responsible to handle this. */
-	static int old_xsize, old_ysize;
-	static int cur_xsize, cur_ysize;
-
-	ignore_signals();
-	if (get_terminal_size(1, &old_xsize, &old_ysize)) return;
-	while (1) {
-		if (get_terminal_size(1, &cur_xsize, &cur_ysize)) return;
-		if ((old_xsize != cur_xsize) || (old_ysize != cur_ysize)) {
-			int wr;
-			old_xsize = cur_xsize;
-			old_ysize = cur_ysize;
-			EINTRLOOP(wr, (int)write(winch_pipe[1], "x", 1));
-			/* Resizing may take some time. So don't send a flood
-		     of requests?! */
-			_sleep2(2*WINCH_SLEEPTIME);
-		}
-		else
-			_sleep2(WINCH_SLEEPTIME);
-	}
-}
-
-static void winch(void *s)
-{
-	unsigned char c;
-	while (can_read(winch_pipe[0])) {
-		int rd;
-		EINTRLOOP(rd, (int)read(winch_pipe[0], &c, 1));
-		if (rd != 1) break;
-	}
-	((void (*)(void))s)();
-}
-
-void handle_terminal_resize(int fd, void (*fn)(void))
-{
-	if (!is_xterm()) return;
-	if (!winch_thread_running) {
-		if (c_pipe(winch_pipe) < 0) return;
-		winch_thread_running = 1;
-		if (_beginthread(winch_thread, NULL, 0x32000, NULL) == -1) {
-		}
-	}
-	set_handlers(winch_pipe[0], winch, NULL, fn);
-}
-
-void unhandle_terminal_resize(int fd)
-{
-	if (!winch_thread_running) return;
-	set_handlers(winch_pipe[0], NULL, NULL, NULL);
-}
-
-int get_terminal_size(int fd, int *x, int *y)
-{
-	if (is_xterm()) {
-#ifdef X2
-		int arc;
-		struct winsize win;
-
-		/* fd = STDIN_FILENO; */
-		arc = ptioctl(1, TIOCGWINSZ, &win);
-		if (arc) {
-			*x = 80;
-			*y = 24;
-			return 0;
-		}
-		*y = win.ws_row;
-		*x = win.ws_col;
-		goto set_default;
-#else
-		*x = 80; *y = 24;
-		return 0;
-#endif
-	} else {
-		int a[2] = { 0, 0 };
-		_scrsize(a);
-		*x = a[0];
-		*y = a[1];
-#ifdef X2
-		set_default:
-#endif
-		if (*x == 0) {
-			*x = get_e("COLUMNS");
-			if (*x == 0) *x = 80;
-		}
-		if (*y == 0) {
-			*y = get_e("LINES");
-			if (*y == 0) *y = 24;
-		}
-	}
-	return 0;
-}
-
 #endif
 
 
@@ -675,7 +698,7 @@ static void fd_unlock(void)
 	_fmutex_release(&fd_mutex);
 }
 
-#elif defined(HAVE_PTHREADS)
+#elif defined(HAVE_PTHREADS) && !defined(OPENVMS)
 
 static void fd_lock(void)
 {
@@ -714,9 +737,7 @@ static void new_fd_bin(int fd)
 #endif
 }
 
-/* Pipe */
-
-#if !defined(OPENVMS) && !defined(DOS)
+#if !defined(OPENVMS)
 
 void set_nonblock(int fd)
 {
@@ -730,81 +751,108 @@ void set_nonblock(int fd)
 #endif
 }
 
-int c_pipe(int *fd)
+#endif
+
+static int cleanup_fds(void)
 {
-	int r;
-	fd_lock();
-	EINTRLOOP(r, pipe(fd));
-	if (!r) new_fd_bin(fd[0]), new_fd_bin(fd[1]);
-	fd_unlock();
-	return r;
+#ifdef ENFILE
+	if (errno == ENFILE) return abort_background_connections();
+#endif
+#ifdef EMFILE
+	if (errno == EMFILE) return abort_background_connections();
+#endif
+	return 0;
 }
 
-#endif
+/* Pipe */
+
+int c_pipe(int fd[2])
+{
+	int r;
+	do {
+		fd_lock();
+		EINTRLOOP(r, pipe(fd));
+		if (!r) new_fd_bin(fd[0]), new_fd_bin(fd[1]);
+		fd_unlock();
+	} while (r == -1 && cleanup_fds());
+	return r;
+}
 
 int c_dup(int oh)
 {
 	int h;
-	fd_lock();
-	EINTRLOOP(h, dup(oh));
-	if (h != -1) new_fd_cloexec(h);
-	fd_unlock();
+	do {
+		fd_lock();
+		EINTRLOOP(h, dup(oh));
+		if (h != -1) new_fd_cloexec(h);
+		fd_unlock();
+	} while (h == -1 && cleanup_fds());
 	return h;
 }
 
 int c_socket(int d, int t, int p)
 {
 	int h;
-	fd_lock();
-	EINTRLOOP(h, socket(d, t, p));
-	if (h != -1) new_fd_cloexec(h);
-	fd_unlock();
+	do {
+		fd_lock();
+		EINTRLOOP(h, socket(d, t, p));
+		if (h != -1) new_fd_cloexec(h);
+		fd_unlock();
+	} while (h == -1 && cleanup_fds());
 	return h;
 }
 
-int c_accept(int h, struct sockaddr *addr, socklen_t *addrlen)
+int c_accept(int sh, struct sockaddr *addr, socklen_t *addrlen)
 {
-	int rh;
-	fd_lock();
-	EINTRLOOP(rh, accept(h, addr, addrlen));
-	if (rh != -1) new_fd_cloexec(rh);
-	fd_unlock();
-	return rh;
+	int h;
+	do {
+		fd_lock();
+		EINTRLOOP(h, accept(sh, addr, addrlen));
+		if (h != -1) new_fd_cloexec(h);
+		fd_unlock();
+	} while (h == -1 && cleanup_fds());
+	return h;
 }
 
 int c_open(unsigned char *path, int flags)
 {
 	int h;
-	fd_lock();
-	EINTRLOOP(h, open(cast_const_char path, flags));
-	if (h != -1) new_fd_bin(h);
-	fd_unlock();
+	do {
+		fd_lock();
+		EINTRLOOP(h, open(cast_const_char path, flags));
+		if (h != -1) new_fd_bin(h);
+		fd_unlock();
+	} while (h == -1 && cleanup_fds());
 	return h;
 }
 
 int c_open3(unsigned char *path, int flags, int mode)
 {
 	int h;
-	fd_lock();
-	EINTRLOOP(h, open(cast_const_char path, flags, mode));
-	if (h != -1) new_fd_bin(h);
-	fd_unlock();
+	do {
+		fd_lock();
+		EINTRLOOP(h, open(cast_const_char path, flags, mode));
+		if (h != -1) new_fd_bin(h);
+		fd_unlock();
+	} while (h == -1 && cleanup_fds());
 	return h;
 }
 
 DIR *c_opendir(unsigned char *path)
 {
 	DIR *d;
-	fd_lock();
-	ENULLLOOP(d, opendir(cast_const_char path));
+	do {
+		fd_lock();
+		ENULLLOOP(d, opendir(cast_const_char path));
 #ifdef HAVE_DIRFD
-	if (d) {
-		int h;
-		EINTRLOOP(h, dirfd(d));
-		if (h != -1) new_fd_cloexec(h);
-	}
+		if (d) {
+			int h;
+			EINTRLOOP(h, dirfd(d));
+			if (h != -1) new_fd_cloexec(h);
+		}
 #endif
-	fd_unlock();
+		fd_unlock();
+	} while (!d && cleanup_fds());
 	return d;
 }
 
@@ -903,7 +951,7 @@ int is_xterm(void)
 	return xt;
 }
 
-#elif defined(BEOS) || defined(ATHEOS) || defined(DOS)
+#elif defined(BEOS) || defined(ATHEOS) || defined(DOS) || defined(HAIKU)
 
 int is_xterm(void)
 {
@@ -939,9 +987,9 @@ void close_fork_tty(void)
 	struct k_conn *k;
 	struct list_head *lk;
 	int rs;
+	reinit_child(1);
 #ifndef NO_SIGNAL_HANDLERS
 	EINTRLOOP(rs, close(signal_pipe[0]));
-	EINTRLOOP(rs, close(signal_pipe[1]));
 #endif
 #ifdef G
 	if (drv && drv->after_fork) drv->after_fork();
@@ -976,7 +1024,7 @@ void get_path_to_exe(void)
 	static unsigned char path[4096];
 	r = GetModuleFileNameA(NULL, cast_char path, sizeof path);
 	if (r <= 0 || r >= sizeof path) {
-		path_to_exe = g_argv[0];
+		path_to_exe = cast_uchar g_argv[0];
 		return;
 	}
 	path_to_exe = path;
@@ -990,7 +1038,7 @@ void get_path_to_exe(void)
 	   the quotation marks will be present in g_argv[0] ... and will
 	   prevent executing it */
 	static unsigned char path[270];
-	path_to_exe = g_argv[0];
+	path_to_exe = cast_uchar g_argv[0];
 	if (!os2_pib) return;
 	if (DosQueryModuleName(os2_pib->pib_hmte, sizeof path, path)) return;
 	path_to_exe = path;
@@ -1047,7 +1095,7 @@ int os_get_system_name(unsigned char *buffer)
 
 void get_path_to_exe(void)
 {
-	path_to_exe = g_argv[0];
+	path_to_exe = cast_uchar g_argv[0];
 }
 
 #endif
@@ -1182,7 +1230,7 @@ unsigned char *os_conv_to_external_path(unsigned char *file, unsigned char *prog
 			 * Unix shell hates backslash and Windows applications
 			 * accept '/'
 			 */
-			if (*p == '\\') add_to_str(&newstr, &newstrl, cast_uchar "/");
+			if (*p == '\\') add_chr_to_str(&newstr, &newstrl, '/');
 			else add_chr_to_str(&newstr, &newstrl, *p);
 		}
 		return newstr;
@@ -1271,7 +1319,7 @@ unsigned char *os_fixup_external_program(unsigned char *prog)
 #endif
 
 
-#if defined(UNIX) || defined(INTERIX) || defined(BEOS) || defined(RISCOS) || defined(ATHEOS) || defined(SPAD) || defined(OPENVMS) || defined(DOS)
+#if defined(UNIX) || defined(INTERIX) || defined(BEOS) || defined(RISCOS) || defined(ATHEOS) || defined(SPAD) || defined(OPENVMS) || defined(DOS) || defined(HAIKU)
 
 #if defined(BEOS) && defined(HAVE_SETPGID)
 
@@ -1739,7 +1787,7 @@ int resize_window(int x, int y)
 	unsigned char buffer[1024];
 	unsigned char *x1;
 	if (is_xterm()) return -1;
-	if (get_terminal_size(1, &old_x, &old_y)) return -1;
+	get_terminal_size(&old_x, &old_y);
 	x1 = cast_uchar GETSHELL;
 	if (!x1) x1 = cast_uchar DEFAULT_SHELL;
 	if (!is_winnt()) {
@@ -1756,7 +1804,7 @@ int resize_window(int x, int y)
 	   mode. */
 	/* I'm not sure what's the behavior on WinNT 4. Anybody wants to test?
 	   */
-		if (!fullscreen && !get_terminal_size(1, &new_x, &new_y) && (new_x != x || new_y != y)) {
+		if (!fullscreen && (get_terminal_size(&new_x, &new_y), (new_x != x || new_y != y))) {
 			fullscreen = 1;
 #ifdef __CYGWIN__
 			keybd_event(VK_MENU, 0x38, 0, 0);
@@ -1767,7 +1815,8 @@ int resize_window(int x, int y)
 			if (y != 25) call_resize(x1, 80, 25);
 			else call_resize(x1, 80, 50);
 			call_resize(x1, x, y);
-			if (get_terminal_size(1, &new_x, &new_y) || new_x != x || new_y != y) call_resize(x1, old_x, old_y);
+			get_terminal_size(&new_x, &new_y);
+			if (new_x != x || new_y != y) call_resize(x1, old_x, old_y);
 #ifdef __CYGWIN__
 			keybd_event(VK_MENU, 0x38, 0, 0);
 			keybd_event(VK_RETURN, 0x1c, 0, 0);
@@ -2038,7 +2087,7 @@ static uint32 bgat(void *t)
 
 #endif
 
-#if defined(UNIX) || defined(OS2) || defined(WIN) || defined(INTERIX) || defined(RISCOS) || defined(ATHEOS) || defined(SPAD)
+#if defined(UNIX) || defined(OS2) || defined(WIN) || defined(INTERIX) || defined(RISCOS) || defined(ATHEOS) || defined(SPAD) || defined(HAIKU)
 
 void terminate_osdep(void)
 {
@@ -2238,7 +2287,7 @@ static void mouse_remove_pointer(void)
 	A_DECL(NOPTRRECT, pa);
 	static int x = -1, y = -1;
 	static tcount c = -1;
-	if (x == -1 || y == -1 || (c != resize_count)) get_terminal_size(1, &x, &y), c = resize_count;
+	if (x == -1 || y == -1 || (c != resize_count)) get_terminal_size(&x, &y), c = resize_count;
 	pa->row = 0;
 	pa->col = 0;
 	pa->cRow = y - 1;
@@ -2411,38 +2460,50 @@ void done_draw(void)
 int vms_thread_high_priority = 0;
 #endif
 
+#ifndef OPENVMS
 static unsigned thread_count = 0;
+#endif
 
 static inline void reset_thread_count(void)
 {
+#ifndef OPENVMS
 	fd_lock();
 	thread_count = 0;
 	fd_unlock();
+#endif
 }
 
 static void inc_thread_count(void)
 {
+#ifndef OPENVMS
 	fd_lock();
 	thread_count++;
 	fd_unlock();
+#endif
 }
 
 static void dec_thread_count(void)
 {
+#ifndef OPENVMS
 	fd_lock();
 	if (!thread_count)
-		internal("thread_count underflow");
+		internal_error("thread_count underflow");
 	thread_count--;
 	fd_unlock();
+#endif
 }
 
 static inline unsigned get_thread_count(void)
 {
+#ifndef OPENVMS
 	unsigned val;
 	fd_lock();
 	val = thread_count;
 	fd_unlock();
 	return val;
+#else
+	return 0;
+#endif
 }
 
 static void *bgpt(void *t)
@@ -2797,7 +2858,7 @@ void add_gpm_version(unsigned char **s, int *l)
 #ifdef HAVE_GPM_GETLIBVERSION
 	add_to_str(s, l, cast_uchar " (");
 	add_to_str(s, l, cast_uchar Gpm_GetLibVersion(NULL));
-	add_to_str(s, l, cast_uchar ")");
+	add_chr_to_str(s, l, ')');
 #endif
 }
 
@@ -2836,6 +2897,14 @@ int get_system_env(void)
 	return 0;
 }
 
+#elif defined(HAIKU)
+
+int get_system_env(void)
+{
+	if (!getenv("SSH_CONNECTION")) return ENV_BE;
+	return 0;
+}
+
 #elif defined(WIN)
 
 int get_system_env(void)
@@ -2895,6 +2964,8 @@ unsigned char *links_xterm(void)
 	if (!(xterm = cast_uchar getenv("LINKS_XTERM"))) {
 #ifdef OPENVMS
 		xterm = cast_uchar "CREATE /TERMINAL /WAIT";
+#elif defined(HAIKU)
+		xterm = cast_uchar "Terminal";
 #else
 		xterm = cast_uchar "xterm -e";
 #endif
@@ -2955,7 +3026,7 @@ static int open_in_new_interix(struct terminal *term, unsigned char *exe, unsign
 }
 #endif
 
-#ifdef BEOS
+#if defined(BEOS) || defined(HAIKU)
 static int open_in_new_be(struct terminal *term, unsigned char *exe, unsigned char *param)
 {
 	exec_new_links(term, cast_uchar "Terminal", exe, param);
@@ -3008,7 +3079,7 @@ static const struct {
 #ifdef INTERIX
 	{ENV_INTERIX, open_in_new_interix, TEXT_(T_WINDOW), TEXT_(T_HK_WINDOW)},
 #endif
-#ifdef BEOS
+#if defined(BEOS) || defined(HAIKU)
 	{ENV_BE, open_in_new_be, TEXT_(T_BEOS_TERMINAL), TEXT_(T_HK_BEOS_TERMINAL)},
 #endif
 #ifdef G
@@ -3125,7 +3196,7 @@ void os_detach_console(void)
 #endif
 		EINTRLOOP(rp, fork());
 		if (!rp) {
-			reinit_child();
+			reinit_child(0);
 #if defined(HAVE_PTHREADS)
 			reset_thread_count();
 #endif
@@ -3294,6 +3365,10 @@ int os_default_charset(void)
 	ULONG os2_cp[1];
 	ULONG size = 0;
 	int rc;
+
+	if (is_xterm())
+		return 0;
+
 	rc = DosQueryCp(sizeof(os2_cp), os2_cp, &size);
 	if ((!rc || rc == ERROR_CPLIST_TOO_SMALL) && size >= sizeof(ULONG)) {
 		unsigned char a[8];
@@ -3313,3 +3388,28 @@ int os_default_charset(void)
 }
 
 #endif
+
+
+void os_report_error_va(const char *caption, const char *msg, va_list l)
+{
+#ifdef OS2
+	char msg_buffer[OS_REPORT_ERROR_BUFFER];
+	vsnprintf(msg_buffer, sizeof msg_buffer, msg, l);
+	if (os2_init_pm()) return;
+	WinMessageBox(HWND_DESKTOP, NULLHANDLE, msg_buffer, caption, 0, MB_CANCEL | MB_ERROR | MB_APPLMODAL | MB_MOVEABLE);
+	os2_exit_pm();
+#endif
+#ifdef WIN
+	char msg_buffer[OS_REPORT_ERROR_BUFFER];
+	vsnprintf(msg_buffer, sizeof msg_buffer, msg, l);
+	MessageBox(NULL, msg_buffer, caption, MB_OK | MB_ICONEXCLAMATION);
+#endif
+}
+
+void os_report_error(const char *caption, const char *msg, ...)
+{
+	va_list l;
+	va_start(l, msg);
+	os_report_error_va(caption, msg, l);
+	va_end(l);
+}
