@@ -200,7 +200,7 @@ void cls_redraw_all_terminals(void)
 
 void draw_to_window(struct window *win, void (*fn)(struct terminal *term, void *), void *data)
 {
-	struct terminal *term = win->term;
+	struct terminal * volatile term = win->term;	/* volatile because of setjmp */
 	struct window *w;
 	struct list_head *lw;
 	if (!F) {
@@ -220,7 +220,8 @@ void draw_to_window(struct window *win, void (*fn)(struct terminal *term, void *
 	} else {
 		struct rect r1, *r;
 		struct rect_set *s;
-		int i, a;
+		volatile int i;	/* volatile because of setjmp */
+		int a;
 		if (win->list_entry.prev == &term->windows) {
 			pr(fn(term, data)) {};
 			return;
@@ -543,6 +544,12 @@ direct:
 
 #ifdef G
 
+void t_redraw(struct graphics_device *, struct rect *);
+static void t_resize(struct graphics_device *);
+static void t_kbd(struct graphics_device *, int, int);
+static void t_mouse(struct graphics_device *, int, int, int);
+static void t_extra(struct graphics_device *, int, unsigned char *);
+
 static struct term_spec gfx_term = { init_list_1st(NULL) "", 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, init_list_last(NULL) };
 
 struct terminal *init_gfx_term(void (*root_window)(struct window *, struct links_event *, int), unsigned char *cwd, void *info, int len)
@@ -586,6 +593,7 @@ struct terminal *init_gfx_term(void (*root_window)(struct window *, struct links
 	dev->resize_handler = t_resize;
 	dev->keyboard_handler = t_kbd;
 	dev->mouse_handler = t_mouse;
+	dev->extra_handler = t_extra;
 	{
 		int *ptr;
 		struct links_event ev = { EV_INIT, 0, 0, 0 };
@@ -595,7 +603,7 @@ struct terminal *init_gfx_term(void (*root_window)(struct window *, struct links
 		ptr = mem_alloc(sizeof(int) + len);
 		*ptr = len;
 		memcpy(ptr + 1, info, len);
-		ev.b = (long)ptr;
+		ev.b = (my_intptr_t)ptr;
 		root_window(win, &ev, 0);
 		mem_free(ptr);
 	}
@@ -611,7 +619,7 @@ void t_redraw(struct graphics_device *dev, struct rect *r)
 	register_bottom_half(redraw_windows, term);
 }
 
-void t_resize(struct graphics_device *dev)
+static void t_resize(struct graphics_device *dev)
 {
 	struct terminal *term = dev->user_data;
 	struct window *win;
@@ -626,7 +634,7 @@ void t_resize(struct graphics_device *dev)
 	set_clip_area(dev, &dev->size);
 }
 
-void t_kbd(struct graphics_device *dev, int key, int flags)
+static void t_kbd(struct graphics_device *dev, int key, int flags)
 {
 	struct terminal *term = dev->user_data;
 	struct links_event ev = { EV_KBD, 0, 0, 0 };
@@ -655,7 +663,7 @@ void t_kbd(struct graphics_device *dev, int key, int flags)
 	}
 }
 
-void t_mouse(struct graphics_device *dev, int x, int y, int b)
+static void t_mouse(struct graphics_device *dev, int x, int y, int b)
 {
 	struct terminal *term = dev->user_data;
 	struct links_event ev = { EV_MOUSE, 0, 0, 0 };
@@ -683,6 +691,20 @@ void t_mouse(struct graphics_device *dev, int x, int y, int b)
 	next->handler(next, &ev, 0);
 }
 
+static void t_extra(struct graphics_device *dev, int type, unsigned char *str)
+{
+	struct terminal *term = dev->user_data;
+	struct links_event ev = { EV_EXTRA, 0, 0, 0 };
+	struct window *prev;
+
+	ev.x = type;
+	ev.b = (my_intptr_t)str;
+
+	if (list_empty(term->windows)) return;
+	prev = list_struct(term->windows.prev, struct window);
+	prev->handler(prev, &ev, 0);
+}
+
 #endif
 
 static void in_term(void *term_)
@@ -841,9 +863,9 @@ static inline char_t utf8_hack(char_t c)
 {									\
 	add_to_str(&a, &l, cast_uchar "\033[");				\
 	add_num_to_str(&a, &l, (y) + 1 + term->top_margin);		\
-	add_to_str(&a, &l, cast_uchar ";");				\
+	add_chr_to_str(&a, &l, ';');					\
 	add_num_to_str(&a, &l, (x) + 1 + term->left_margin);		\
-	add_to_str(&a, &l, cast_uchar "H");				\
+	add_chr_to_str(&a, &l, 'H');					\
 	n_chars = 0;							\
 }
 
@@ -886,7 +908,7 @@ static inline char_t utf8_hack(char_t c)
 		} else if (getcompcode(attrib & 7) < getcompcode(attrib >> 3 & 7))\
 			add_to_str(&a, &l, cast_uchar ";7");		\
 		if (attrib & 0100) add_to_str(&a, &l, cast_uchar ";1");	\
-		add_to_str(&a, &l, cast_uchar "m");			\
+		add_chr_to_str(&a, &l, 'm');				\
 	}								\
 	if (c >= ' ' && c != 127 && (c != 155 ||			\
 	    (term_charset(term) != utf8_table && cp2u(155, term_charset(term)) != -1))) {\
@@ -904,7 +926,7 @@ static inline char_t utf8_hack(char_t c)
 			r = u2cp(c, 0, 1);				\
 			if (!(r && r[0] >= 32 && r[0] < 127 && !r[1])) r = cast_uchar "*";\
 			add_chr_to_str(&a, &l, r[0]);			\
-			if (cx + 1 < term->x)				\
+			if (term->x - cx > 1)				\
 				add_chr_to_str(&a, &l, 8);		\
 			else						\
 				SETPOS(cx, y);				\
@@ -992,9 +1014,9 @@ static void redraw_screen(struct terminal *term)
 		term->lcy = term->cy;
 		add_to_str(&a, &l, cast_uchar "\033[");
 		add_num_to_str(&a, &l, term->cy + 1 + term->top_margin);
-		add_to_str(&a, &l, cast_uchar ";");
+		add_chr_to_str(&a, &l, ';');
 		add_num_to_str(&a, &l, term->cx + 1 + term->left_margin);
-		add_to_str(&a, &l, cast_uchar "H");
+		add_chr_to_str(&a, &l, 'H');
 	}
 	if (l && term->master) want_draw();
 #ifdef OPENVMS
@@ -1273,7 +1295,7 @@ static void unblock_terminal(void *term_)
 		term->dirty = 0;
 		/*redraw_terminal_cls(term);*/
 #ifdef G
-	} else {
+	} else if (drv->unblock) {
 		drv->unblock(term->dev);
 #endif
 	}
@@ -1325,7 +1347,7 @@ void exec_on_terminal(struct terminal *term, unsigned char *path, unsigned char
 			if (fg == 1) {
 				if (!F) block_itrm(term->fdin);
 #ifdef G
-				else if (drv->block(term->dev)) {
+				else if (drv->block && drv->block(term->dev)) {
 					mem_free(param);
 					if (*delet)
 						EINTRLOOP(rs, unlink(cast_const_char delet));
@@ -1338,7 +1360,7 @@ void exec_on_terminal(struct terminal *term, unsigned char *path, unsigned char
 				if (fg == 1) {
 					if (!F) unblock_itrm(term->fdin);
 #ifdef G
-					else drv->unblock(term->dev);
+					else if (drv->unblock) drv->unblock(term->dev);
 #endif
 				}
 				mem_free(param);
